@@ -6,7 +6,6 @@ from __future__ import annotations
 # Will no longer be needed in Python 3.10.
 
 import abc
-import concurrent.futures
 import contextlib
 import datetime
 import os
@@ -15,6 +14,7 @@ import pathlib
 import queue
 import sys
 import threading
+from concurrent.futures import ThreadPoolExecutor
 from dataclasses import dataclass
 from io import UnsupportedOperation
 from typing import (
@@ -71,103 +71,6 @@ class FileInfo:
 
 
 class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-methods
-    _thread_executor_ = {
-        "nest0": concurrent.futures.ThreadPoolExecutor(
-            min(32, (os.cpu_count() or 1) + 4), thread_name_prefix="UpathExecutor0"
-        ),
-        "nest1": concurrent.futures.ThreadPoolExecutor(
-            10, thread_name_prefix="UpathExecutor1"
-        ),
-    }
-    # Currently there can be two "layers" of threads running during `download_dir`.
-    # In `download_dir`, the download of each file runs in the threads provided
-    # by `nest0`. If a file is large, `GcpBlobUpath` will split the work into chunks
-    # and download each chunk in a thread provided by `nest1`.
-    # We dedicate the two executors mainly so that the second layer of chunk downloads
-    # do not starve for threads.
-
-    @classmethod
-    def _run_in_executor(
-        cls, tasks: Iterable[Tuple[Callable, tuple, dict, str]], description: str
-    ):
-        """
-        This method is used to run multiple I/O jobs concurrently, e.g.
-        uploading/downloading all files in a folder recursively.
-
-        `tasks`: each element is a tuple of (func, args, kwargs, description).
-
-        `description`: description of the entire set of tasks, such as "Downloading directory 'abc'".
-        """
-        if not isinstance(tasks, list):
-            tasks = list(tasks)
-        n_tasks = len(tasks)
-        if not n_tasks:
-            return
-
-        def enqueue(q_tasks, executor, q):
-            while True:
-                try:
-                    func, args, kwargs, desc = q_tasks.get_nowait()
-                except queue.Empty:
-                    break
-                t = executor.submit(func, *args, **kwargs)
-                q.put((t, desc))
-            q.put(None)
-
-        pbar = None
-        if threading.current_thread().name.startswith("UpathExecutor0"):
-            # This is the case when GCP downloads a large file by multiple parts.
-            executor = cls._thread_executor_["nest1"]
-        else:
-            executor = cls._thread_executor_["nest0"]
-            print(description, file=sys.stderr)
-            pbar = tqdm(
-                total=n_tasks,
-                bar_format="{percentage:5.1f}%, {n:.0f}/{total_fmt}, {elapsed} | {desc}",
-            )
-
-        try:
-            q = queue.Queue(executor._max_workers + 4)
-            q_tasks = queue.SimpleQueue()
-            for t in tasks:
-                q_tasks.put(t)
-            task = threading.Thread(target=enqueue, args=(q_tasks, executor, q))
-            task.start()
-
-            try:
-                while True:
-                    z = q.get()
-                    if z is None:
-                        break
-                    t, desc = z
-                    if pbar:
-                        pbar.set_description_str(desc)
-                        pbar.update(0.5)
-                    try:
-                        yield t.result()
-                    except Exception:
-                        while True:
-                            try:
-                                _ = q_tasks.get_nowait()
-                            except queue.Empty:
-                                break
-                        while True:
-                            z = q.get()
-                            if z is None:
-                                break
-                            t, desc = z
-                            t.cancel()
-                            # This may not succeed, but there isn't a good way to
-                            # guarantee cancellation here.
-                        raise
-                    if pbar:
-                        pbar.update(0.5)
-            finally:
-                task.join()
-        finally:
-            if pbar:
-                pbar.close()
-
     @classmethod
     def register_read_write_byte_format(cls, serde: Type[ByteSerializer], name: str):
         """
@@ -211,11 +114,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         setattr(cls, f"write_{name}", _write)
         setattr(cls, f"read_{name}", _read)
 
-    def __init__(self, *pathsegments: str):
+    def __init__(
+        self, *pathsegments: str, thread_pool_executors: List[ThreadPoolExecutor] = None
+    ):
         """`Upath` is the base class for a client to a blob store,
         including local file system as a special case.
 
-        `*pathsegments`: analogous to the input to `pathlib.Path`.
+        `*pathsegments`: analogous to the input to `pathleib.Path`.
         The first segment may or may not start with `/`; it makes
         no difference. The path constructed with `*pathsegments`
         is always "absolute" under a known "root".
@@ -229,8 +134,15 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         For Azure blob store, the root is that in a "container".
         For AWS and GCP blob stores, the root is that in a "bucket".
 
+        `thread_pool_executors`: some operations may use threads. If there are
+        a large number of Upath instances active at the same time, the number of
+        threads could be large. You may pass in two thread-pool-executors to use,
+        hence controlling the total number of threads created by this object.
+        This parameter consists of two *separate* executors---don't pass in
+        a single executor twice.
+
         Subclasses for cloud blob stores may need to add additional parameters
-        representing, e.g. credentials, container/bucket name, etc.
+        representing, e.g., container/bucket name, etc.
         """
 
         self._path = os.path.normpath(
@@ -239,6 +151,24 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         # The path is always "absolute" starting with '/'.
         # It does not have a trailing `/` unless the path is just `/` itself.
 
+        if thread_pool_executors is None:
+            self._thread_pools = []
+        elif thread_pool_executors == []:
+            self._thread_pools = thread_pool_executors
+        else:
+            assert len(thread_pool_executors) == 2
+            e0, e1 = thread_pool_executors
+            e0._thread_name_prefix = "UpathExecutor0"
+            e1._thread_name_prefix = "UpathExecutor1"
+            self._thread_pools = thread_pool_executors
+
+    def __getstate__(self):
+        return (self._path,)
+
+    def __setstate__(self, data):
+        self._path = data[0]
+        self._thread_pools = []
+
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}('{self._path}')"
         # Subclass may want to customize this method to add more info,
@@ -283,8 +213,114 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         """
         return self.joinpath(key)
 
+    @property
+    def _thread_pool_executors(self):
+        if not self._thread_pools:
+            self._thread_pools.append(
+                ThreadPoolExecutor(thread_name_prefix="UpathExecutor0")
+            )
+            self._thread_pools.append(
+                ThreadPoolExecutor(10, thread_name_prefix="UpathExecutor1")
+            )
+        return self._thread_pools
+        # Currently there can be two "layers" of threads running during `download_dir`.
+        # In `download_dir`, the download of each file runs in the threads provided
+        # by `nest0`. If a file is large, `GcsBlobUpath` will split the work into chunks
+        # and download each chunk in a thread provided by `nest1`.
+        # We dedicate the two executors mainly so that the second layer of chunk downloads
+        # do not starve for threads.
+
+    def _run_in_executor(
+        self,
+        tasks: Iterable[Tuple[Callable, tuple, dict, str]],
+        description: str,
+    ):
+        """
+        This method is used to run multiple I/O jobs concurrently, e.g.
+        uploading/downloading all files in a folder recursively.
+
+        `tasks`: each element is a tuple of (func, args, kwargs, description).
+
+        `description`: description of the entire set of tasks, such as "Downloading directory 'abc'".
+            Note: if `description` is a false value, progress printouts will be suppressed.
+        """
+        if not isinstance(tasks, list):
+            tasks = list(tasks)
+        n_tasks = len(tasks)
+        if not n_tasks:
+            return
+
+        def enqueue(q_tasks, executor, q):
+            while True:
+                try:
+                    func, args, kwargs, desc = q_tasks.get_nowait()
+                except queue.Empty:
+                    break
+                t = executor.submit(func, *args, **kwargs)
+                q.put((t, desc))
+            q.put(None)
+
+        pbar = None
+        if threading.current_thread().name.startswith("UpathExecutor0"):
+            # This is the case when GCP downloads a large file by multiple parts.
+            executor = self._thread_pool_executors[1]
+        else:
+            executor = self._thread_pool_executors[0]
+            if description:
+                print(description, file=sys.stderr)
+                pbar = tqdm(
+                    total=n_tasks,
+                    bar_format="{percentage:5.1f}%, {n:.0f}/{total_fmt}, {elapsed} | {desc}",
+                )
+
+        try:
+            q = queue.Queue(executor._max_workers + 4)
+            q_tasks = queue.SimpleQueue()
+            for t in tasks:
+                q_tasks.put(t)
+            task = threading.Thread(target=enqueue, args=(q_tasks, executor, q))
+            task.start()
+
+            try:
+                while True:
+                    z = q.get()
+                    if z is None:
+                        break
+                    t, desc = z
+                    if pbar:
+                        pbar.set_description_str(desc)
+                        pbar.update(0.5)
+                    try:
+                        yield t.result()
+                    except Exception:
+                        while True:
+                            try:
+                                _ = q_tasks.get_nowait()
+                            except queue.Empty:
+                                break
+                        while True:
+                            z = q.get()
+                            if z is None:
+                                break
+                            t, desc = z
+                            t.cancel()
+                            # This may not succeed, but there isn't a good way to
+                            # guarantee cancellation here.
+                        raise
+                    if pbar:
+                        pbar.update(0.5)
+            finally:
+                task.join()
+        finally:
+            if pbar:
+                pbar.close()
+
     def copy_dir(
-        self, target: str, *, overwrite: bool = False, desc: str = None
+        self,
+        target: str,
+        *,
+        overwrite: bool = False,
+        quiet: bool = False,
     ) -> int:
         """Analogous to `copy_file`.
 
@@ -304,10 +340,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                     extra,
                 )
 
+        if quiet:
+            desc = False
+        else:
+            desc = f"Copying from {self} into {target_}"
+
         n = 0
-        for _ in self._run_in_executor(
-            foo(), desc or f"Copying from {self} into {target_}"
-        ):
+        for _ in self._run_in_executor(foo(), desc):
             n += 1
         return n
 
@@ -356,7 +395,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         return self.is_file() or self.is_dir()
 
     def export_dir(
-        self, target: Upath, *, overwrite: bool = False, desc: str = None
+        self,
+        target: Upath,
+        *,
+        overwrite: bool = False,
+        quiet: bool = False,
     ) -> int:
         """Copy the content of the current directory recursively
         to the specified `target`, which is typically in another store.
@@ -388,10 +431,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                     extra,
                 )
 
+        if quiet:
+            desc = False
+        else:
+            desc = f"Exporting from {self!r} into {target!r}"
+
         n = 0
-        for _ in self._run_in_executor(
-            foo(), desc or f"Exporting from {self!r} into {target!r}"
-        ):
+        for _ in self._run_in_executor(foo(), desc):
             n += 1
         return n
 
@@ -425,7 +471,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         raise NotImplementedError
 
     def import_dir(
-        self, source: Upath, *, overwrite: bool = False, desc: str = None
+        self,
+        source: Upath,
+        *,
+        overwrite: bool = False,
+        quiet: bool = False,
     ) -> int:
         """Analogous to `export_dir`."""
 
@@ -440,10 +490,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                     extra,
                 )
 
+        if quiet:
+            desc = False
+        else:
+            desc = f"Importing from {source!r} into {self!r}"
+
         n = 0
-        for _ in self._run_in_executor(
-            foo(), desc or f"Importing from {source!r} into {self!r}"
-        ):
+        for _ in self._run_in_executor(foo(), desc):
             n += 1
         return n
 
@@ -600,9 +653,7 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         # Refer to https://docs.python.org/3/library/functions.html#open
         return self.read_bytes().decode(encoding="utf-8", errors="strict")
 
-    # TODO: rename 'remove' to 'delete'?
-
-    def remove_dir(self, *, desc: str = None) -> int:
+    def remove_dir(self, *, quiet: bool = True) -> int:
         """Remove the directory pointed to by `self`,
         along with all its contents, recursively.
 
@@ -616,8 +667,13 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
             for p in self.riterdir():
                 yield p.remove_file, [], {}, str(p.path.relative_to(self.path))
 
+        if quiet:
+            desc = False
+        else:
+            desc = f"Removing {self!r}"
+
         n = 0
-        for _ in self._run_in_executor(foo(), desc or f"Removing {self!r}"):
+        for _ in self._run_in_executor(foo(), desc):
             n += 1
         return n
 
@@ -632,7 +688,11 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         raise NotImplementedError
 
     def rename_dir(
-        self: T, target: str, *, overwrite: bool = False, desc: str = None
+        self: T,
+        target: str,
+        *,
+        overwrite: bool = False,
+        quiet: bool = False,
     ) -> T:
         """Analogous to `rename_file`.
 
@@ -660,9 +720,12 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
                     extra,
                 )
 
-        for _ in self._run_in_executor(
-            foo(), desc or f"Renaming {self!r} to {target_!r}"
-        ):
+        if quiet:
+            desc = False
+        else:
+            desc = f"Renaming {self!r} to {target_!r}"
+
+        for _ in self._run_in_executor(foo(), desc):
             pass
         return target_
 
@@ -706,7 +769,7 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         """
         raise NotImplementedError
 
-    def rmrf(self) -> int:
+    def rmrf(self, *, quiet: bool = True) -> int:
         """Analogous to `rm -rf`. Remove the file or dir pointed to
         by `self`.
 
@@ -729,7 +792,7 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         else:
             n = 1
         try:
-            m = self.remove_dir()
+            m = self.remove_dir(quiet=quiet)
         except FileNotFoundError:
             m = 0
         return n + m
@@ -760,7 +823,7 @@ class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-metho
         Subclass needs to reimplement this method
         if its `__init__` expects additional args.
         """
-        return self.__class__(*paths)
+        return self.__class__(*paths, thread_pool_executors=self._thread_pools)
 
     # def with_stem(self: T, stem: str) -> T:
     #     # Available in Python 3.9+.
